/**
 * @file
 * Attaches behavior for the Edit module.
 *
 * Everything happens asynchronously, to allow for:
 *   - dynamically rendered contextual links
 *   - asynchronously retrieved (and cached) per-field in-place editing metadata
 *   - asynchronous setup of in-place editable field and "Quick edit" link
 *
 * To achieve this, there are several queues:
 *   - fieldsMetadataQueue: fields whose metadata still needs to be fetched.
 *   - fieldsAvailableQueue: queue of fields whose metadata is known, and for
 *     which it has been confirmed that the user has permission to edit them.
 *     However, FieldModels will only be created for them once there's a
 *     contextual link for their entity: when it's possible to initiate editing.
 *   - contextualLinksQueue: queue of contextual links on entities for which it
 *     is not yet known whether the user has permission to edit at >=1 of them.
 */

(function ($, _, Backbone, Drupal, drupalSettings) {

"use strict";

var options = $.extend(drupalSettings.edit,
  // Merge strings on top of drupalSettings so that they are not mutable.
  {
    strings: {
      quickEdit: Drupal.t('Quick edit')
    }
  }
);

/**
 * Tracks fields without metadata. Contains objects with the following keys:
 *   - DOM el
 *   - String fieldID
 *   - String entityID
 */
var fieldsMetadataQueue = [];

/**
 * Tracks fields ready for use. Contains objects with the following keys:
 *   - DOM el
 *   - String fieldID
 *   - String entityID
 */
var fieldsAvailableQueue = [];

/**
 * Tracks contextual links on entities. Contains objects with the following
 * keys:
 *   - String entityID
 *   - DOM el
 *   - DOM region
 */
var contextualLinksQueue = [];

/**
 * Tracks how many instances exist for each unique entity. Contains key-value
 * pairs:
 * - String entityID
 * - Number count
 */
var entityInstancesTracker = {};

Drupal.behaviors.edit = {
  attach: function (context) {
    // Initialize the Edit app once per page load.
    $('body').once('edit-init', initEdit);

    // Process each entity element: identical entities that appear multiple
    // times will get a numeric identifier, starting at 0.
    $(context).find('[data-edit-entity-id]').once('edit').each(function (index, entityElement) {
      var entityID = entityElement.getAttribute('data-edit-entity-id');
      if (!entityInstancesTracker.hasOwnProperty(entityID)) {
        entityInstancesTracker[entityID] = 0;
      }
      else {
        entityInstancesTracker[entityID]++;
      }

      // Set the calculated entity instance ID for this element.
      var entityInstanceID = entityInstancesTracker[entityID];
      entityElement.setAttribute('data-edit-entity-instance-id', entityInstanceID);
    });

    // Detect contextual links on entities annotated by Edit; queue these to be
    // processed.
    $(context).find('.contextual-links').once('edit-contextual').each(function (index, contextualLinkElement) {
      var $region = $(contextualLinkElement).closest('.contextual-links-region');
      if ($region.is('[data-edit-entity-id]')) {
        var contextualLink = {
          entityID: $region.attr('data-edit-entity-id'),
          entityInstanceID: $region.attr('data-edit-entity-instance-id'),
          el: contextualLinkElement,
          region: $region[0]
        };
        // Queue contextual link to be set up later.
        contextualLinksQueue.push(contextualLink);
      }
    });

    // Process each field element: queue to be used or to fetch metadata.
    // When a field is being rerendered after editing, it will be processed
    // immediately. New fields will be unable to be processed immediately, but
    // will instead be queued to have their metadata fetched, which occurs below
    // in fetchMissingMetaData().
    $(context).find('[data-edit-field-id]').once('edit').each(function (index, fieldElement) {
      processField(fieldElement);
    });

    // Fetch metadata for any fields that are queued to retrieve it.
    fetchMissingMetadata(function (fieldElementsWithFreshMetadata) {
      // Metadata has been fetched, reprocess fields whose metadata was missing.
      _.each(fieldElementsWithFreshMetadata, processField);

      // Metadata has been fetched, try to set up more contextual links now.
      contextualLinksQueue = _.filter(contextualLinksQueue, function (contextualLink) {
        return !initializeEntityContextualLink(contextualLink);
      });
    });
  },
  detach: function (context, settings, trigger) {
    if (trigger === 'unload') {
      deleteContainedModelsAndQueues($(context));
    }
  }
};

Drupal.edit = {
  // A Drupal.edit.AppView instance.
  app: null,

  collections: {
    // All in-place editable entities (Drupal.edit.EntityModel) on the page.
    entities: null,
    // All in-place editable fields (Drupal.edit.FieldModel) on the page.
    fields: null
  },

  // In-place editors will register themselves in this object.
  editors: {},

  // Per-field metadata that indicates whether in-place editing is allowed,
  // which in-place editor should be used, etc.
  metadata: {
    has: function (fieldID) { return _.has(this.data, fieldID); },
    add: function (fieldID, metadata) { this.data[fieldID] = metadata; },
    get: function (fieldID, key) {
      return (key === undefined) ? this.data[fieldID] : this.data[fieldID][key];
    },
    intersection: function (fieldIDs) { return _.intersection(fieldIDs, _.keys(this.data)); },
    // Contains the actual metadata, keyed by field ID.
    data: {}
  }
};

/**
 * Extracts the entity ID from a field ID.
 *
 * @param String fieldID
 *   A field ID: a string of the format
 *   `<entity type>/<id>/<field name>/<language>/<view mode>`.
 * @return String
 *   An entity ID: a string of the format `<entity type>/<id>`.
 */
function extractEntityID (fieldID) {
  return fieldID.split('/').slice(0, 2).join('/');
}

/**
 * Initialize the Edit app.
 *
 * @param DOM bodyElement
 *   This document's body element.
 */
function initEdit (bodyElement) {
  Drupal.edit.collections.entities = new Drupal.edit.EntityCollection();
  Drupal.edit.collections.fields = new Drupal.edit.FieldCollection();

  // Instantiate AppModel (application state) and AppView, which is the
  // controller of the whole in-place editing experience.
  Drupal.edit.app = new Drupal.edit.AppView({
    el: bodyElement,
    model: new Drupal.edit.AppModel(),
    entitiesCollection: Drupal.edit.collections.entities,
    fieldsCollection: Drupal.edit.collections.fields
  });
}

/**
 * Fetch the field's metadata; queue or initialize it (if EntityModel exists).
 *
 * @param DOM fieldElement
 *   A Drupal Field API field's DOM element with a data-edit-field-id attribute.
 */
function processField (fieldElement) {
  var metadata = Drupal.edit.metadata;
  var fieldID = fieldElement.getAttribute('data-edit-field-id');
  var entityID = extractEntityID(fieldID);
  // Figure out the instance ID by looking at the ancestor [data-edit-entity-id]
  // element's data-edit-entity-instance-id attribute.
  var entityElementSelector = '[data-edit-entity-id="' + entityID + '"]';
  var entityElement = $(fieldElement).closest(entityElementSelector);
  // In the case of a full entity view page, the entity title is rendered
  // outside of "the entity DOM node": it's rendered as the page title. So in
  // this case, we must find the entity in the mandatory "content" region.
  if (entityElement.length === 0) {
    entityElement = $('.region-content').find(entityElementSelector);
  }
  var entityInstanceID = entityElement
    .get(0)
    .getAttribute('data-edit-entity-instance-id');

  // Early-return if metadata for this field is missing.
  if (!metadata.has(fieldID)) {
    fieldsMetadataQueue.push({
      el: fieldElement,
      fieldID: fieldID,
      entityID: entityID,
      entityInstanceID: entityInstanceID
    });
    return;
  }
  // Early-return if the user is not allowed to in-place edit this field.
  if (metadata.get(fieldID, 'access') !== true) {
    return;
  }

  // If an EntityModel for this field already exists (and hence also a "Quick
  // edit" contextual link), then initialize it immediately.
  if (Drupal.edit.collections.entities.findWhere({ entityID: entityID, entityInstanceID: entityInstanceID })) {
    initializeField(fieldElement, fieldID, entityID, entityInstanceID);
  }
  // Otherwise: queue the field. It is now available to be set up when its
  // corresponding entity becomes in-place editable.
  else {
    fieldsAvailableQueue.push({ el: fieldElement, fieldID: fieldID, entityID: entityID, entityInstanceID: entityInstanceID });
  }
}

/**
 * Initialize a field; create FieldModel.
 *
 * @param DOM fieldElement
 *   The field's DOM element.
 * @param String fieldID
 *   The field's ID.
 * @param String entityID
 *   The field's entity's ID.
 * @param String entityInstanceID
 *   The field's entity's instance ID.
 */
function initializeField (fieldElement, fieldID, entityID, entityInstanceID) {
  var entity = Drupal.edit.collections.entities.findWhere({
    entityID: entityID,
    entityInstanceID: entityInstanceID
  });

  $(fieldElement).addClass('edit-field');

  // The FieldModel stores the state of an in-place editable entity field.
  var field = new Drupal.edit.FieldModel({
    el: fieldElement,
    fieldID: fieldID,
    id: fieldID + '[' + entity.get('entityInstanceID') + ']',
    entity: entity,
    metadata: Drupal.edit.metadata.get(fieldID),
    acceptStateChange: _.bind(Drupal.edit.app.acceptEditorStateChange, Drupal.edit.app)
  });

  // Track all fields on the page.
  Drupal.edit.collections.fields.add(field);
}

/**
 * Fetches metadata for fields whose metadata is missing.
 *
 * Fields whose metadata is missing are tracked at fieldsMetadataQueue.
 *
 * @param Function callback
 *   A callback function that receives field elements whose metadata will just
 *   have been fetched.
 */
function fetchMissingMetadata (callback) {
  if (fieldsMetadataQueue.length) {
    var fieldIDs = _.pluck(fieldsMetadataQueue, 'fieldID');
    var fieldElementsWithoutMetadata = _.pluck(fieldsMetadataQueue, 'el');
    var entityIDs = _.uniq(_.pluck(fieldsMetadataQueue, 'entityID'), true);
    // Ensure we only request entityIDs for which we don't have metadata yet.
    entityIDs = _.difference(entityIDs, _.keys(Drupal.edit.metadata.data));
    fieldsMetadataQueue = [];

    $.ajax({
      url: drupalSettings.edit.metadataURL,
      type: 'POST',
      data: {
        'fields[]': fieldIDs,
        'entities[]': entityIDs
      },
      dataType: 'json',
      success: function(results) {
        // Store the metadata.
        _.each(results, function (fieldMetadata, fieldID) {
          Drupal.edit.metadata.add(fieldID, fieldMetadata);
        });

        callback(fieldElementsWithoutMetadata);
      }
    });
  }
}

/**
 * Loads missing in-place editor's attachments (JavaScript and CSS files).
 *
 * Missing in-place editors are those whose fields are actively being used on
 * the page but don't have
 *
 * @param Function callback
 *   Callback function to be called when the missing in-place editors (if any)
 *   have been inserted into the DOM. i.e. they may still be loading.
 */
function loadMissingEditors (callback) {
  var loadedEditors = _.keys(Drupal.edit.editors);
  var missingEditors = [];
  Drupal.edit.collections.fields.each(function (fieldModel) {
    var metadata = Drupal.edit.metadata.get(fieldModel.get('fieldID'));
    if (metadata.access && _.indexOf(loadedEditors, metadata.editor) === -1) {
      missingEditors.push(metadata.editor);
    }
  });
  missingEditors = _.uniq(missingEditors);
  if (missingEditors.length === 0) {
    callback();
  }

  // @todo Simplify this once https://drupal.org/node/1533366 lands.
  // @see https://drupal.org/node/2029999.
  var id = 'edit-load-editors';
  // Create a temporary element to be able to use Drupal.ajax.
  var $el = $('<div id="' + id + '" class="element-hidden"></div>').appendTo('body');
  // Create a Drupal.ajax instance to load the form.
  var loadEditorsAjax = new Drupal.ajax(id, $el, {
    url: drupalSettings.edit.attachmentsURL,
    event: 'edit-internal.edit',
    submit: { 'editors[]': missingEditors },
    // No progress indicator.
    progress: { type: null }
  });
  // Work-around for https://drupal.org/node/2019481 in Drupal 7.
  loadEditorsAjax.commands = {};
  // The above work-around prevents the prototype implementations from being
  // called, so we must alias any and all of the commands that might be called.
  loadEditorsAjax.commands.settings = Drupal.ajax.prototype.commands.settings;
  // Implement a scoped insert AJAX command: calls the callback after all AJAX
  // command functions have been executed (hence the deferred calling).
  var realInsert = Drupal.ajax.prototype.commands.insert;
  loadEditorsAjax.commands.insert = function (ajax, response, status) {
    _.defer(callback);
    realInsert(ajax, response, status);
    $el.off('edit-internal.edit');
    $el.remove();
  };
  // Trigger the AJAX request, which will should return AJAX commands to insert
  // any missing attachments.
  $el.trigger('edit-internal.edit');
}

/**
 * Attempts to set up a "Quick edit" link and corresponding EntityModel.
 *
 * @param Object contextualLink
 *   An object with the following properties:
 *     - String entityID: an Edit entity identifier, e.g. "node/1" or
 *       "custom_block/5".
 *     - String entityInstanceID: an Edit entity instance identifier, e.g. 0, 1
 *       or n (depending on whether it's the first, second, or n+1st instance of
 *       this entity).
 *     - DOM el: element pointing to the contextual links placeholder for this
 *       entity.
 *     - DOM region: element pointing to the contextual region for this entity.
 * @return Boolean
 *   Returns true when a contextual the given contextual link metadata can be
 *   removed from the queue (either because the contextual link has been set up
 *   or because it is certain that in-place editing is not allowed for any of
 *   its fields).
 *   Returns false otherwise.
 */
function initializeEntityContextualLink (contextualLink) {
  var metadata = Drupal.edit.metadata;
  // Check if the user has permission to edit at least one of them.
  function hasFieldWithPermission (fieldIDs) {
    for (var i = 0; i < fieldIDs.length; i++) {
      var fieldID = fieldIDs[i];
      if (metadata.get(fieldID, 'access') === true) {
        return true;
      }
    }
    return false;
  }

  // Checks if the metadata for all given field IDs exists.
  function allMetadataExists (fieldIDs) {
    return fieldIDs.length === metadata.intersection(fieldIDs).length;
  }

  // Find all fields for this entity instance and collect their field IDs.
  var fields = _.where(fieldsAvailableQueue, {
    entityID: contextualLink.entityID,
    entityInstanceID: contextualLink.entityInstanceID
  });
  var fieldIDs = _.pluck(fields, 'fieldID');

  // No fields found yet.
  if (fieldIDs.length === 0) {
    return false;
  }
  // The entity for the given contextual link contains at least one field that
  // the current user may edit in-place; instantiate EntityModel,
  // EntityDecorationView and ContextualLinkView.
  else if (hasFieldWithPermission(fieldIDs)) {
    var entityModel = new Drupal.edit.EntityModel({
      el: contextualLink.region,
      entityID: contextualLink.entityID,
      entityInstanceID: contextualLink.entityInstanceID,
      id: contextualLink.entityID + '[' + contextualLink.entityInstanceID + ']',
      label: Drupal.edit.metadata.get(contextualLink.entityID, 'label')
    });
    Drupal.edit.collections.entities.add(entityModel);
    // Create an EntityDecorationView associated with the root DOM node of the
    // entity.
    var entityDecorationView = new Drupal.edit.EntityDecorationView({
      el: contextualLink.region,
      model: entityModel
    });
    entityModel.set('entityDecorationView', entityDecorationView);

    // Initialize all queued fields within this entity (creates FieldModels).
    _.each(fields, function (field) {
      initializeField(field.el, field.fieldID, contextualLink.entityID, contextualLink.entityInstanceID);
    });
    fieldsAvailableQueue = _.difference(fieldsAvailableQueue, fields);

    // Initialization should only be called once. Use Underscore's once method
    // to get a one-time use version of the function.
    var initContextualLink = _.once(function () {
      var $links = $(contextualLink.el);
      var contextualLinkView = new Drupal.edit.ContextualLinkView($.extend({
        el: $('<li class="quick-edit"><a href="" role="button" aria-pressed="false"></a></li>').prependTo($links),
        model: entityModel,
        appModel: Drupal.edit.app.model
      }, options));
      entityModel.set('contextualLinkView', contextualLinkView);
    });

    // Set up ContextualLinkView after loading any missing in-place editors.
    loadMissingEditors(initContextualLink);

    return true;
  }
  // There was not at least one field that the current user may edit in-place,
  // even though the metadata for all fields within this entity is available.
  else if (allMetadataExists(fieldIDs)) {
    return true;
  }

  return false;
}

/**
 * Delete models and queue items that are contained within a given context.
 *
 * Deletes any contained EntityModels (plus their associated FieldModels and
 * ContextualLinkView) and FieldModels, as well as the corresponding queues.
 *
 * After EntityModels, FieldModels must also be deleted, because it is possible
 * in Drupal for a field DOM element to exist outside of the entity DOM element,
 * e.g. when viewing the full node, the title of the node is not rendered within
 * the node (the entity) but as the page title.
 *
 * Note: this will not delete an entity that is actively being in-place edited.
 *
 * @param jQuery $context
 *   The context within which to delete.
 */
function deleteContainedModelsAndQueues($context) {
  $context.find('[data-edit-entity-id]').addBack('[data-edit-entity-id]').each(function (index, entityElement) {
    // Delete entity model.
    var entityModel = Drupal.edit.collections.entities.findWhere({el: entityElement});
    if (entityModel) {
      var contextualLinkView = entityModel.get('contextualLinkView');
      contextualLinkView.remove();
      // Remove the EntityDecorationView.
      entityModel.get('entityDecorationView').remove();
      // Destroy the EntityModel; this will also destroy its FieldModels.
      entityModel.destroy();
    }

    // Filter queue.
    function hasOtherRegion (contextualLink) {
      return contextualLink.region !== entityElement;
    }
    contextualLinksQueue = _.filter(contextualLinksQueue, hasOtherRegion);
  });

  $context.find('[data-edit-field-id]').addBack('[data-edit-field-id]').each(function (index, fieldElement) {
    // Delete field models.
    Drupal.edit.collections.fields.chain()
      .filter(function (fieldModel) { return fieldModel.get('el') === fieldElement; })
      .invoke('destroy');

    // Filter queues.
    function hasOtherFieldElement (field) {
      return field.el !== fieldElement;
    }
    fieldsMetadataQueue = _.filter(fieldsMetadataQueue, hasOtherFieldElement);
    fieldsAvailableQueue = _.filter(fieldsAvailableQueue, hasOtherFieldElement);
  });
}

})(jQuery, _, Backbone, Drupal, Drupal.settings);
